@@ -15,7 +15,7 @@
  *
  */
 /*
- *  Copyright (C) 2000-2019 Thomas Habets <thomas@habets.se>
+ *  Copyright (C) 2000-2022 Thomas Habets <thomas@habets.se>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -109,6 +109,13 @@
 
 #if HAVE_SYS_CAPABILITY_H
 #include <sys/capability.h>
+#else
+// It seems that some environments have a libc with cap_init, but do not have
+// the header files. Without the header files we won't have the cap_t struct, so
+// it won't actually work.
+//
+// TODO: Probably this logic should be in configure.ac instead.
+#undef HAVE_CAP_INIT
 #endif
 
 #if HAVE_NET_BPF_H
@@ -116,6 +123,13 @@
 #endif
 #include <pcap.h>
 
+#if defined(HAVE_SECCOMP_H) && defined(HAVE_LIBSECCOMP)
+#define USE_SECCOMP 1
+#include <seccomp.h>
+#else
+#define USE_SECCOMP 0
+#endif
+
 #include "arping.h"
 
 #ifndef ETH_ALEN
@@ -192,6 +206,9 @@ static int finddup = 0;              /* finddup mode. -d */
 static int dupfound = 0;             /* set to 1 if dup found */
 static char lastreplymac[ETH_ALEN];  /* if last different from this then dup */
 
+/* -z to turn on, -Z to turn off. Default is compile time option */
+static int use_seccomp = DEFAULT_SECCOMP;
+
 unsigned int numsent = 0;                   /* packets sent */
 unsigned int numrecvd = 0;                  /* packets received */
 static unsigned int max_replies = UINT_MAX; /* exit after -C replies */
@@ -458,13 +475,80 @@ drop_privileges(const char* drop_group)
 #endif
 }
 
+#if USE_SECCOMP
+static void seccomp_allow(scmp_filter_ctx ctx, const char* name)
+{
+        if (seccomp_rule_add(ctx, SCMP_ACT_ALLOW, seccomp_syscall_resolve_name(name), 0)) {
+                if (verbose) {
+                        fprintf(stderr, "arping: seccomp_rule_add_exact(%s): %s",
+                                name, strerror(errno));
+                }
+        }
+}
+
+static void drop_seccomp(int libnet_fd)
+{
+        //scmp_filter_ctx ctx = seccomp_init(SCMP_ACT_ERRNO(13));
+        scmp_filter_ctx ctx = seccomp_init(SCMP_ACT_KILL);
+        if (!ctx) {
+                perror("seccomp_init()");
+                exit(1);
+        }
+
+        //
+        // Whitelist.
+        //
+
+        // Write to stdout and stderr.
+        if (seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(fstat), 1, SCMP_A0(SCMP_CMP_EQ, STDOUT_FILENO))) {
+                perror("seccomp_rule_add(fstat stdout)");
+                exit(1);
+        }
+        if (seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(write), 1, SCMP_A0(SCMP_CMP_EQ, STDOUT_FILENO))) {
+                perror("seccomp_rule_add(write stdout)");
+                exit(1);
+        }
+        if (seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(write), 1, SCMP_A0(SCMP_CMP_EQ, STDERR_FILENO))) {
+                perror("seccomp_rule_add(write stderr)");
+                exit(1);
+        }
+
+        // Libnet.
+        if (seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(ioctl), 1, SCMP_A0(SCMP_CMP_EQ, libnet_fd))) {
+                perror("seccomp_rule_add(ioctl libnet)");
+                exit(1);
+        }
+        if (seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(sendto), 1, SCMP_A0(SCMP_CMP_EQ, libnet_fd))) {
+                perror("seccomp_rule_add(sendto libnet)");
+                exit(1);
+        }
+
+        // Other.
+        seccomp_allow(ctx, "select");
+        seccomp_allow(ctx, "pselect6");
+        seccomp_allow(ctx, "newfstatat");
+        seccomp_allow(ctx, "exit_group");
+        seccomp_allow(ctx, "rt_sigreturn");
+
+        // Load.
+        if (seccomp_load(ctx)) {
+                perror("seccomp_load()");
+                exit(1);
+        }
+        seccomp_release(ctx);
+        if (verbose > 1) {
+                printf("arping: Successfully applied seccomp policy\n");
+        }
+}
+#endif
+
 /**
  * drop even more privileges, where possible.
  *
  * After all setup is done and main loop is about to start.
  */
 static void
-drop_more_privileges()
+drop_more_privileges(int libnet_fd)
 {
 #ifdef HAVE_PLEDGE
         if (pledge("stdio tty", "")) {
@@ -475,8 +559,44 @@ drop_more_privileges()
                 printf("arping: Successfully pledged\n");
         }
 #endif
+#if USE_SECCOMP
+        if (use_seccomp) {
+                drop_seccomp(libnet_fd);
+        }
+#endif
 }
 
+/**
+ * Check for buggy libpcap version
+ * https://github.com/the-tcpdump-group/libpcap/issues/461
+ *
+ * This list is not authoritative.
+ *
+ * If it's over matching wrong the only impact *should* be that too many packets
+ * come through the BPF filter.
+ *
+ * If it's under-matching then replies with dot1p nonzero will not be seen.
+ */
+static int
+bug_pcap_vlan()
+{
+        const char* bad[] = {
+                // Broken in 1.7.0.
+                "libpcap version 1.7",
+                "libpcap version 1.8",
+                "libpcap version 1.9.0",
+                // Fixed in 1.9.1.
+                // Fixed in 1.10 and up.
+                NULL,
+        };
+        const char* v = pcap_lib_version();
+        for (const char** t = bad; *t; t++) {
+                if (!strncmp(v, *t, strlen(*t))) {
+                        return 1;
+                }
+        }
+        return 0;
+}
 
 /**
  * Do pcap_open_live(), except by using the pcap_create() interface
@@ -626,17 +746,48 @@ strip_newline(char* s) {
 
 /**
  * Init libnet with specified ifname. Destroy if already inited.
- * If this function retries with different parameter it will preserve
- * the original error message and print that.
+ *
+ * Libnet usually needs init before we have searched for the real
+ * interface. In that case, first we just give a NULL pointer as the
+ * interface. But libnet sometimes fails to find an interface (no idea
+ * why), so then we try to use "lo" and "lo0" explicitly.
+ *
+ * If even loopback fails, then it'll preserve the original error
+ * message.
+ *
  * Call with recursive=0.
  */
 void
-do_libnet_init(const char *ifname, int recursive)
+do_libnet_init(const char *inifname, int recursive)
 {
+        const char* ifname = inifname;
+        int last = 0;
+        switch (recursive) {
+        case 0:
+                break;
+        case 1:
+                ifname = "lo"; // E.g. Linux.
+                break;
+        case 2:
+                ifname = "lo0"; // E.g. OpenBSD.
+                break;
+        default:
+                last = 1;
+                break;
+        }
+
+        // If we're given an interface name then always use that.
+        // No need to be recursive about it.
+        if (inifname != NULL && recursive == 0) {
+                ifname = inifname;
+                last = 1;
+        }
+
 	char ebuf[LIBNET_ERRBUF_SIZE];
         ebuf[0] = 0;
 	if (verbose > 1) {
-                printf("arping: libnet_init(%s)\n", ifname ? ifname : "<null>");
+                printf("arping: trying libnet_init(LIBNET_LINK, %s)\n",
+                       ifname ? ifname : "<null>");
 	}
 	if (libnet) {
 		/* Probably going to switch interface from temp to real. */
@@ -644,28 +795,31 @@ do_libnet_init(const char *ifname, int recursive)
 		libnet = 0;
 	}
 
-        /* Try libnet_init() even though we aren't root. We may have
+        /* Try libnet_init() even though we maybe aren't root. We may have
          * a capability or something. */
 	if (!(libnet = libnet_init(LIBNET_LINK,
 				   (char*)ifname,
 				   ebuf))) {
                 strip_newline(ebuf);
-                if (!ifname) {
-                        /* Sometimes libnet guesses an interface that it then
-                         * can't use. Work around that by attempting to
-                         * use "lo". */
-                        do_libnet_init("lo", 1);
+                if (verbose) {
+                        fprintf(stderr,
+                                "arping: libnet_init(LIBNET_LINK, %s): %s\n",
+                                ifname ? ifname : "<null>",
+                                *ebuf ? ebuf : "<no error message>");
+                }
+                if (!last) {
+                        do_libnet_init(ifname, recursive+1);
                         if (libnet != NULL) {
                                 return;
                         }
-                } else if (recursive) {
-                        /* Continue original execution to get that
-                         * error message. */
-                        return;
                 }
-                fprintf(stderr, "arping: libnet_init(LIBNET_LINK, %s): %s\n",
-                        ifname ? ifname : "<null>",
-                        *ebuf ? ebuf : "<no error message>");
+                if (!verbose) {
+                        // Prevent double-print when verbose.
+                        fprintf(stderr,
+                                "arping: libnet_init(LIBNET_LINK, %s): %s\n",
+                                ifname ? ifname : "<null>",
+                                *ebuf ? ebuf : "<no error message>");
+                }
                 if (getuid() && geteuid()) {
                         fprintf(stderr,
                                 "arping: you may need to run as root\n");
@@ -790,9 +944,16 @@ extended_usage()
 	       "    -U     Send unsolicited ARP.\n"
 	       "    -v     Verbose output. Use twice for more messages.\n"
                "    -V num 802.1Q tag to add. Defaults to no VLAN tag.\n"
-               "    -w sec Specify a timeout before ping exits regardless of how"
-               " many\npackets have been sent or received.\n"
-               "    -W sec Time to wait between pings.\n");
+               "    -w sec Specify a timeout before ping exits regardless of how many\n"
+               "           packets have been sent or received.\n"
+               "    -W sec Time to wait between pings.\n"
+#if USE_SECCOMP
+               "    -z     Enable seccomp%s\n"
+               "    -Z     Disable seccomp%s\n",
+               DEFAULT_SECCOMP ? " (default)" : "",
+               DEFAULT_SECCOMP ? "" : " (default)"
+#endif
+               );
         printf("Report bugs to: thomas@habets.se\n"
                "Arping home page: <http://www.habets.pp.se/synscan/>\n"
                "Development repo: http://github.com/ThomasHabets/arping\n");
@@ -806,7 +967,7 @@ standard_usage()
 {
 	printf("ARPing %s, by Thomas Habets <thomas@habets.se>\n",
 	       version);
-        printf("usage: arping [ -0aAbdDeFpPqrRuUv ] [ -w <sec> ] "
+        printf("usage: arping [ -0aAbdDeFpPqrRuUv%s ] [ -w <sec> ] "
                "[ -W <sec> ] "
                "[ -S <host/ip> ]\n"
                "              "
@@ -817,7 +978,7 @@ standard_usage()
                " [ -g <group> ]\n"
                "              "
                "[ -V <vlan> ] [ -Q <priority> ] "
-               "<host/ip/MAC | -B>\n");
+               "<host/ip/MAC | -B>\n", USE_SECCOMP ? "zZ" : "");
 }
 
 /**
@@ -836,6 +997,19 @@ usage(int ret)
 	exit(ret);
 }
 
+/**
+ *
+ */
+static void
+print_library_versions()
+{
+        fprintf(stderr, "arping: %s\n", pcap_lib_version());
+        if (bug_pcap_vlan()) {
+                fprintf(stderr, "arping: bug: https://github.com/the-tcpdump-group/libpcap/issues/461\n");
+        }
+        fprintf(stderr, "arping: %s\n", libnet_version());
+}
+
 /**
  * Check to see if it looks somewhat like a MAC address.
  *
@@ -1125,7 +1299,7 @@ pingip_send()
                                           send_reply ? ARPOP_REPLY : ARPOP_REQUEST,
 					  srcmac,
 					  (uint8_t*)&srcip,
-					  unsolicited ? (uint8_t*)ethxmas : (uint8_t*)ethnull,
+					  unsolicited ? (uint8_t*)ethxmas : (send_reply ? dstmac : (uint8_t*)ethnull),
 					  (uint8_t*)&dstip,
 					  (uint8_t*)padding,
 					  sizeof padding,
@@ -1188,7 +1362,7 @@ void
 pingip_recv(const char *unused, struct pcap_pkthdr *h, const char * const packet)
 {
         const unsigned char *pkt_srcmac;
-        const struct libnet_802_1q_hdr *veth;
+        const struct libnet_802_1q_hdr *veth = NULL;
 	struct libnet_802_3_hdr *heth;
 	struct libnet_arp_hdr *harp;
         struct timespec arrival;
@@ -1201,28 +1375,57 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, const char * const packet
         getclock(&arrival);
 
 	if (vlan_tag >= 0) {
+                if (h->caplen < LIBNET_802_1Q_H + LIBNET_ARP_H + 2*(ETH_ALEN + 4)) {
+                        return;
+                }
 		veth = (void*)packet;
 		harp = (void*)((char*)veth + LIBNET_802_1Q_H);
 		pkt_srcmac = veth->vlan_shost;
 	} else {
-                // Short packet.
                 if (h->caplen < LIBNET_ETH_H + LIBNET_ARP_H + 2*(ETH_ALEN + 4)) {
                         return;
                 }
-
 		heth = (void*)packet;
 		harp = (void*)((char*)heth + LIBNET_ETH_H);
 		pkt_srcmac = heth->_802_3_shost;
-                // Wrong length of hardware address.
-                if (harp->ar_hln != ETH_ALEN) {
-                        return;
-                }
+        }
+        if (verbose > 3) {
+                printf("arping: ... good length\n");
+        }
 
-                // Wrong length of protocol address.
-                if (harp->ar_pln != 4) {
+        if (veth) {
+                if (veth->vlan_tpi != htons(0x8100)) {
                         return;
                 }
-	}
+                if (verbose > 3) {
+                        printf("arping: ... is dot1q\n");
+                }
+                if ((veth->vlan_priority_c_vid & 0xfff) == vlan_tag) {
+                        return;
+                }
+                if (verbose > 3) {
+                        printf("arping: ... right VLAN\n");
+                }
+        }
+
+        // Not checking ethertype because in theory this could be used for
+        // Ethernet II.
+
+        // Wrong length of hardware address.
+        if (harp->ar_hln != ETH_ALEN) {
+                return;
+        }
+        if (verbose > 3) {
+                printf("arping: ... L2 addr len is correct\n");
+        }
+
+        // Wrong length of protocol address.
+        if (harp->ar_pln != 4) {
+                return;
+        }
+        if (verbose > 3) {
+                printf("arping: ... L3 addr len is correct\n");
+        }
 
         // ARP reply.
         if (htons(harp->ar_op) != ARPOP_REPLY) {
@@ -1365,7 +1568,7 @@ pingmac_recv(const char* unused, struct pcap_pkthdr *h, uint8_t *packet)
 {
         const unsigned char *pkt_dstmac;
         const unsigned char *pkt_srcmac;
-        const struct libnet_802_1q_hdr *veth;
+        const struct libnet_802_1q_hdr *veth = NULL;
 	struct libnet_802_3_hdr *heth;
 	struct libnet_ipv4_hdr *hip;
 	struct libnet_icmpv4_hdr *hicmp;
@@ -1379,18 +1582,45 @@ pingmac_recv(const char* unused, struct pcap_pkthdr *h, uint8_t *packet)
         getclock(&arrival);
 
         if (vlan_tag >= 0) {
+                if (h->caplen < LIBNET_ETH_H + LIBNET_IPV4_H + LIBNET_ICMPV4_H) {
+                        return;
+                }
                 veth = (void*)packet;
                 hip = (void*)((char*)veth + LIBNET_802_1Q_H);
                 hicmp = (void*)((char*)hip + LIBNET_IPV4_H);
                 pkt_srcmac = veth->vlan_shost;
                 pkt_dstmac = veth->vlan_dhost;
         } else {
+                if (h->caplen < LIBNET_ETH_H + LIBNET_ARP_H + LIBNET_ICMPV4_H) {
+                        return;
+                }
                 heth = (void*)packet;
                 hip = (void*)((char*)heth + LIBNET_ETH_H);
                 hicmp = (void*)((char*)hip + LIBNET_IPV4_H);
                 pkt_srcmac = heth->_802_3_shost;
                 pkt_dstmac = heth->_802_3_dhost;
         }
+        if (verbose > 3) {
+                printf("arping: ... good length\n");
+        }
+
+        if (veth) {
+                if (veth->vlan_tpi != htons(0x8100)) {
+                        return;
+                }
+                if (verbose > 3) {
+                        printf("arping: ... is dot1q\n");
+                }
+                if ((veth->vlan_priority_c_vid & 0xfff) == vlan_tag) {
+                        return;
+                }
+                if (verbose > 3) {
+                        printf("arping: ... right VLAN\n");
+                }
+        }
+
+        // Not checking ethertype because in theory this could be used for
+        // Ethernet II.
 
         // Dest MAC must be me.
         if (memcmp(pkt_dstmac, srcmac, ETH_ALEN)) {
@@ -1641,7 +1871,6 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
 			       break;
 		       }
 	       }
-
 	       if (trydispatch) {
 		       int ret;
                        if (0 > (ret = pcap_dispatch(pcap, -1,
@@ -1722,7 +1951,7 @@ arping_main(int argc, char **argv)
 	memcpy(dstmac, ethxmas, ETH_ALEN);
 
         while (EOF != (c = getopt(argc, argv,
-                                  "0aAbBC:c:dDeFg:hi:I:m:pPqQ:rRs:S:t:T:uUvV:w:W:"))) {
+                                  "0aAbBC:c:dDeFg:hi:I:m:pPqQ:rRs:S:t:T:uUvV:w:W:zZ"))) {
 		switch(c) {
 		case '0':
 			srcip_opt = "0.0.0.0";
@@ -1850,6 +2079,12 @@ arping_main(int argc, char **argv)
                 case 'W':
                         packetwait = (unsigned)(1000000.0 * atof(optarg));
                         break;
+                case 'z':
+                        use_seccomp = 1;
+                        break;
+                case 'Z':
+                        use_seccomp = 0;
+                        break;
 		default:
 			usage(1);
 		}
@@ -1862,6 +2097,10 @@ arping_main(int argc, char **argv)
                 exit(1);
         }
 
+        if (verbose > 2) {
+                print_library_versions();
+        }
+
         // Generate random payload suffix for MAC pings, to be able to
         // differentiate from unrelated ping replies.
         if (payload_suffix_size < 0) {
@@ -2022,6 +2261,9 @@ arping_main(int argc, char **argv)
 	 * libnet init (may be done already for resolving)
 	 */
         do_libnet_init(ifname, 0);
+        if (verbose > 1) {
+                printf("arping: libnet_getfd(): %d\n", libnet_getfd(libnet));
+        }
 
 	/*
 	 * Make sure dstip and parm like eachother
@@ -2143,9 +2385,14 @@ arping_main(int argc, char **argv)
 	}
 #endif
 
+        if (vlan_tag >= 0 && bug_pcap_vlan() && verbose) {
+                fprintf(stderr,
+                        "arping: Working around bug in libpcap 1.7-1.9.0.\n");
+        }
+
 	if (mode == PINGIP) {
 		/* FIXME: better filter with addresses? */
-                if (vlan_tag >= 0) {
+                if (vlan_tag >= 0 && !bug_pcap_vlan()) {
                         snprintf(bpf_filter, sizeof(bpf_filter),
                                  "vlan %u and arp", vlan_tag);
                 } else {
@@ -2158,7 +2405,7 @@ arping_main(int argc, char **argv)
 		}
 	} else { /* ping mac */
 		/* FIXME: better filter with addresses? */
-                if (vlan_tag >= 0) {
+                if (vlan_tag >= 0 && !bug_pcap_vlan()) {
                         snprintf(bpf_filter, sizeof(bpf_filter),
                                  "vlan %u and icmp", vlan_tag);
                 } else {
@@ -2208,7 +2455,7 @@ arping_main(int argc, char **argv)
                        format_mac(srcmac, buf, sizeof(buf)));
 	}
 
-        drop_more_privileges();
+        drop_more_privileges(libnet_getfd(libnet));
 
 	if (display == NORMAL) {
 		printf("ARPING %s\n", parm);
